# :coding: utf-8

import logging
import os
import sys
import tempfile
import textwrap

import click
import wiz
import wiz.config

import qip
import qip._logging
from qip import __version__

# Initiate logging handler to display potential warning when fetching config.
qip._logging.initiate()

#: Retrieve configuration mapping to initialize default values.
_CONFIG = wiz.config.fetch()

#: Click default context for all commands.
CONTEXT_SETTINGS = dict(
    max_content_width=90,
    help_option_names=["-h", "--help"],
)


@click.group(
    context_settings=CONTEXT_SETTINGS
)
@click.version_option(version=__version__)
@click.option(
    "-v", "--verbosity",
    help="Set the logging output verbosity.",
    type=click.Choice(qip._logging.LEVEL_MAPPING.keys()),
    default="info"
)
def main(**kwargs):
    """Quarantine Installer for Python."""
    qip._logging.initiate(console_level=kwargs["verbosity"])


@main.command(
    name="install",
    help=textwrap.dedent(
        """
        Install one or several packages.
    
        Command example:

        \b
        >>> qip install .
        >>> qip install /path/to/foo/
        >>> qip install foo
        >>> qip install foo bar
        >>> qip install "foo==0.1.0"
        >>> qip install "foo >= 7, < 8"
        >>> qip install "git@gitlab:rnd/foo.git"
        >>> qip install "git@gitlab:rnd/foo.git@0.1.0"
        >>> qip install "git@gitlab:rnd/foo.git@dev"
        >>> qip install foo -p "python==3.6.*"
        """
    ),
    short_help="Install one or several packages.",
    context_settings=CONTEXT_SETTINGS
)
@click.option(
    "-o", "--output-path",
    help=(
        "Destination for the package installation data. Default is "
        "'<TEMPORARY_FOLDER>/qip/packages'"
    ),
    type=click.Path(),
    metavar="PATH",
    default=(
        _CONFIG.get("qip", {}).get(
            "packages_output", os.path.join(
                tempfile.gettempdir(), "qip", "packages"
            )
        )
    ),
)
@click.option(
    "-d", "--definition-path",
    help=(
        "Destination for the Wiz definitions extracted. Default is "
        "'<TEMPORARY_FOLDER>/qip/definitions'"
    ),
    type=click.Path(),
    metavar="PATH",
    default=(
        _CONFIG.get("qip", {}).get(
            "definitions_output", os.path.join(
                tempfile.gettempdir(), "qip", "definitions"
            )
        )
    ),
)
@click.option(
    "-u", "--update",
    help=(
        "Include additional variants from existing Wiz definitions, using "
        "definitions previously exported in the same definitions output path."
    ),
    is_flag=True,
    default=False
)
@click.option(
    "-N", "--no-dependencies",
    help="Install the specified package(s) without dependencies.",
    is_flag=True,
    default=False
)
@click.option(
    "-f/-s", "--overwrite-installed/--skip-installed",
    help=(
        "Indicate whether packages already installed should be overwritten "
        "or skipped. By default, a user confirmation will be required."
    ),
    default=None
)
@click.option(
    "-e", "--editable",
    help=(
        "Install a project in editable mode (i.e. setuptools \"develop mode\") "
        "from a local project path or a VCS url."
    ),
    is_flag=True,
    default=False
)
@click.option(
    "-I", "--ignore-registries",
    help=(
         "Ignore Wiz registries when determining whether a package "
         "should be skipped or updated."
    ),
    is_flag=True,
    default=False
)
@click.option(
    "-p", "--python",
    help=(
        "Target a specific Python version via a Wiz request or a path to a "
        "Python executable."
    ),
    show_default=True,
    metavar="TARGET",
    default=_CONFIG.get("qip", {}).get("python_target", sys.executable)
)
@click.option(
    "-R", "--continue-on-error",
    help="Continue installation process when a package fails to install.",
    is_flag=True,
    default=False
)
@click.argument(
    "requests",
    nargs=-1,
    required=True
)
def install(**kwargs):
    """Install one or several packages."""
    logger = logging.getLogger(__name__ + ".install")

    output_path = kwargs["output_path"]
    definition_path = kwargs["definition_path"]

    registry_paths = []

    if not kwargs["ignore_registries"]:
        registry_paths += wiz.registry.get_defaults()

    if kwargs["update"]:
        registry_paths += [definition_path]

    try:
        success = qip.install(
            kwargs["requests"], output_path,
            definition_path=definition_path,
            overwrite=kwargs["overwrite_installed"],
            no_dependencies=kwargs["no_dependencies"],
            editable_mode=kwargs["editable"],
            python_target=kwargs["python"],
            registry_paths=registry_paths,
            update_existing_definitions=kwargs["update"],
            continue_on_error=kwargs["continue_on_error"],
        )
    except RuntimeError as error:
        raise click.exceptions.ClickException(
            "Impossible to resume installation process:\n\n{}".format(error)
        )

    if not success:
        raise click.exceptions.ClickException("No packages installed.")

    logger.info("Package output directory: {!r}".format(output_path))
    logger.info("Definition output directory: {!r}".format(definition_path))



# :coding: utf-8

import argparse
import json
import re
import sys

import pkg_resources

#: Compiled regular expression to detect request with extra option.
REQUEST_PATTERN = re.compile(r"(.*)\[(\w*)]")

#: Metadata file containing the top-level Python module name.
TOP_LEVEL_METADATA_FILE = "top_level.txt"


def display_package_mapping(name):
    """Display Python package information mapping from selected *name*.

    :term:`Pip` Python API is being used to fetch useful information about a
    package within an environment. The information is displayed as a
    :term:`JSON` encoded mapping so that it can easily be retrieved via a
    subprocess.

    :param name: Python package name.

    :return: None

    Example::

        >>> display_package_mapping("foo")

        {
            "package": {
                "installed_version": "1.0.0",
                "key": "foo",
                "package_name": "Foo",
                "module_name": "foo"
            },
            "requirements": [
                "bim<3,>=2",
                "baz",
            ]
        }


    """
    # Identify and extract 'extra' label from input *name* if necessary.
    match = REQUEST_PATTERN.match(name)
    labels = []

    if match is not None:
        labels = [match.group(2)]
        name = match.group(1)

    # Identify selected package.
    package = pkg_resources.get_distribution(name.lower())

    # Module name should be fetched from the metadata, but we will use the
    # package name if we cannot access the top_level.txt file.
    module_name = package.project_name
    if package.has_metadata(TOP_LEVEL_METADATA_FILE):
        module_name = package.get_metadata(TOP_LEVEL_METADATA_FILE).strip()

    result = {
        "package": {
            "key": package.key,
            "package_name": package.project_name,
            "module_name": module_name,
            "installed_version": package.version,
        },
        "requirements": [
            requirement.key + str(requirement.specifier)
            for requirement in package.requires(extras=labels)
        ]
    }

    print(json.dumps(result, sort_keys=True, indent=4))


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        prog="package-info",
        description="Query information about an installed Python package",
    )
    parser.add_argument("name", help="Python package name to query.")
    namespace = parser.parse_args()

    sys.exit(display_package_mapping(namespace.name))



# :coding: utf-8

import collections
import getpass
import logging
import logging.config
import os
import sys
import tempfile

import coloredlogs
import wiz.filesystem

# Configure custom colors for messages displayed in the console.
coloredlogs.DEFAULT_LEVEL_STYLES = {
    "info": {"color": "cyan"},
    "error": {"color": "red"},
    "critical": {"color": "red"},
    "warning": {"color": "yellow"}
}

#: Available levels with corresponding labels.
LEVEL_MAPPING = collections.OrderedDict([
    ("debug", logging.DEBUG),
    ("info", logging.INFO),
    ("warning", logging.WARNING),
    ("error", logging.ERROR),
])

#: Output path for files exported by default 'file' handler.
PATH = os.path.join(tempfile.gettempdir(), "qip", "logs")


def initiate(console_level="info"):
    """Initiate logger configuration.

    :param console_level: Initialize the logging level for the console handler
        if possible. Default is "info".

    """
    # Ensure that default output path exists.
    wiz.filesystem.ensure_directory(PATH)

    logging.config.dictConfig({
        "version": 1,
        "root": {
            "handlers": ["console", "file"],
            "level": logging.DEBUG
        },
        "formatters": {
            "standard": {
                "class": "coloredlogs.ColoredFormatter",
                "format": "%(message)s"
            },
            "detailed": {
                "class": "logging.Formatter",
                "format": "%(asctime)s - %(levelname)s - %(name)s - %(message)s"
            }
        },
        "handlers": {
            "console": {
                "class": "logging.StreamHandler",
                "formatter": "standard",
                "stream": "ext://sys.stdout",
                "level": LEVEL_MAPPING[console_level]
            },
            "file": {
                "class": "logging.handlers.RotatingFileHandler",
                "formatter": "detailed",
                "level": logging.INFO,
                "filename": os.path.join(
                    PATH, "{}.log".format(getpass.getuser())
                ),
                "maxBytes": 10485760,
                "backupCount": 20,
            }
        }
    })

    # Formatter class cannot be initiate via config in Python 2.7.
    if sys.version_info[0] < 3:
        coloredlogs.install(
            fmt="%(message)s",
            stream=sys.stdout,
            level=console_level.upper(),
        )



# :coding: utf-8

import logging
import json
import os
import re

import wiz.filesystem

import qip.command
import qip.environ
import qip.system

#: Compiled regular expression to detect git input.
GIT_PATTERN = re.compile(r"^git@[\w._-]+:")

#: Compiled regular expression to detect request with extra option.
EXTRA_REQUEST_PATTERN = re.compile(r"(?:.*)\s*\[(.+)]")

#: Path to the Python package info script.
PACKAGE_INFO_SCRIPT = os.path.join(
    os.path.dirname(__file__), "package_data", "package_info.py"
)


def install(request, path, context_mapping, cache_path, editable_mode=False):
    """Install package in *path* from *request*.

    :param request: package to be installed. A request can be one of::

            "/path/to/foo/"
            "."
            "foo"
            "foo==0.1.0"
            "foo >= 7, < 8"
            "git@gitlab:rnd/foo.git"
            "git@gitlab:rnd/foo.git@0.1.0"
            "git@gitlab:rnd/foo.git@dev"

    :param path: path to install Python packages to.

    :param context_mapping: contain environment mapping and python mapping, as
        returned from :func:`qip.fetch_context_mapping`.

    :param cache_path: Temporary directory for the pip cache.

    :param editable_mode: install in editable mode. Default is False.

    :raise RuntimeError: if :term:`Pip` fails to install Python package.

    :raise ValueError: if the Python package name can not be extracted from
        *request*.

    :return: mapping with information about the package gathered from the
        environment. It should be in the form of::

            {
                "identifier": "Foo-0.1.0",
                "request": "foo >= 0.1.0, < 1",
                "name": "Foo",
                "key": "foo",
                "version": "0.1.0",
                "description": "This is a Python package",
                "location": "/path/to/source",
                "target": "Foo/Foo-0.1.0-py27-centos7",
                "python": {
                    "identifier": "2.7",
                    "request": "python >= 2.7, < 2.8",
                    "library-path": "lib/python2.8/site-packages"
                },
                "system": {
                    "platform": "linux",
                    "arch": "x86_64",
                    "os": {
                        "name": "centos",
                        "major_version": 7
                    }
                },
                "requirements": [
                    "bim<3,>=2"
                ]
            }

    """
    logger = logging.getLogger(__name__ + ".install")

    if GIT_PATTERN.match(request) is not None:
        request = "git+ssh://" + request.replace(":", "/")

    logger.debug("Installing '{}'...".format(request))
    result = qip.command.execute(
        "python -m pip install "
        "--ignore-installed "
        "--no-deps "
        "--prefix {destination} "
        "--no-warn-script-location "
        "--disable-pip-version-check "
        "--cache-dir {cache_dir} "
        "{editable_mode}"
        "'{requirement}'".format(
            editable_mode="-e " if editable_mode else "",
            destination=path,
            requirement=request,
            cache_dir=cache_path
        ),
        context_mapping["environ"]
    )

    match_name = re.search("(?<=Installing collected packages: ).*", result)
    if match_name is None:
        raise ValueError(
            "Package name could not be extracted from '{}'.".format(request)
        )
    name = match_name.group().strip()

    extra_keywords = []

    matched_extra = EXTRA_REQUEST_PATTERN.match(request)
    if matched_extra:
        extra_keywords = matched_extra.group(1).split(",")
        extra_keywords = sorted(key.strip() for key in extra_keywords)
        extra_keywords = [key for key in extra_keywords if len(key)]

    mapping = fetch_mapping_from_environ(
        name, context_mapping,
        extra_keywords=extra_keywords
    )
    mapping["request"] = request
    return mapping


def fetch_mapping_from_environ(name, context_mapping, extra_keywords=None):
    """Return a mapping with information about the Python package *name*.

    :param name: Python package name.

    :param context_mapping: contain environment mapping and python mapping, as
        returned from :func:`qip.fetch_context_mapping`.

    :param extra_keywords: List of :term:`extra requirement keywords
        <extras_require>` if required. Default is None.

    :return: mapping with information about the package gathered from the
        environment. It should be in the form of::

            {
                "identifier": "Foo-0.1.0",
                "name": "Foo",
                "key": "foo",
                "version": "0.1.0",
                "description": "This is a Python package",
                "location": "/path/to/source",
                "target": "Foo/Foo-0.1.0-py27-centos7",
                "python": {
                    "identifier": "2.7",
                    "request": "python >= 2.7, < 2.8",
                    "library-path": "lib/python2.8/site-packages"
                },
                "system": {
                    "platform": "linux",
                    "arch": "x86_64",
                    "os": {
                        "name": "centos",
                        "major_version": 7
                    }
                },
                "requirements": [
                    "bim<3,>=2"
                ]
            }

    """
    logger = logging.getLogger(__name__ + ".fetch_mapping_from_environ")

    # Extract package information and its dependency.
    dependency_mapping = extract_dependency_mapping(
        name, context_mapping["environ"],
        extra_keywords=extra_keywords
    )

    # Run pip show command to find extra information from extended metadata.
    metadata = qip.command.execute(
        "python -m pip show "
        "--disable-pip-version-check "
        "'{}' -v".format(name),
        context_mapping["environ"],
        quiet=True
    )

    mapping = {
        "identifier": extract_identifier(
            dependency_mapping["package"],
            extra_keywords=extra_keywords
        ),
        "key": extract_key(
            dependency_mapping["package"],
            extra_keywords=extra_keywords
        ),
        "name": dependency_mapping["package"]["package_name"],
        "module_name": dependency_mapping["package"]["module_name"],
        "version": dependency_mapping["package"]["installed_version"],
        "python": context_mapping["python"]
    }

    match_description = re.search("(?<=Summary: ).+", metadata)
    if match_description is not None:
        mapping["description"] = match_description.group().strip()

    match_location = re.search("(?<=Location: ).+", metadata)
    if match_location is not None:
        mapping["location"] = match_location.group().strip()

    if is_system_required(metadata):
        mapping["system"] = qip.system.query()

    command_mapping = extract_command_mapping(
        metadata, extra_keywords=extra_keywords
    )
    if len(command_mapping) > 0:
        mapping["command"] = command_mapping

    if len(dependency_mapping.get("requirements", [])) > 0:
        mapping["requirements"] = dependency_mapping["requirements"]
        logger.debug(
            "Dependencies: {}".format(" ".join(mapping["requirements"]))
        )

    # Add target information to package mapping.
    mapping["target"] = extract_target_path(
        mapping["name"], mapping["identifier"],
        context_mapping["python"]["identifier"],
        os_mapping=mapping.get("system", {}).get("os"),
    )

    return mapping


def extract_dependency_mapping(name, environ_mapping, extra_keywords=None):
    """Return mapping for Python package with all dependency requirements.

    :param name: Python package name.

    :param environ_mapping: mapping of environment variables.

    :param extra_keywords: List of :term:`extra requirement keywords
        <extras_require>` if required. Default is None.

    :return: None if the package *name* cannot be found in dependency mapping,
        otherwise return dependency mapping. A valid mapping should be in the
        form of::

            {
                "package": {
                    "key": "foo",
                    "package_name": "Foo",
                    "module_name": "foo"
                    "installed_version": "0.1.0",
                },
                "requirements": [
                    "bim<3,>=2"
                ]
            }

    """
    identifier = name.lower()
    if extra_keywords is not None and len(extra_keywords):
        identifier += "[{}]".format(",".join(extra_keywords))

    command = "python {script} {identifier}".format(
        script=PACKAGE_INFO_SCRIPT,
        identifier=identifier
    )

    result = qip.command.execute(command, environ_mapping, quiet=True)
    try:
        mapping = json.loads(result)
    except ValueError:
        raise RuntimeError(
            "Impossible to fetch installed package for '{}'".format(identifier)
        )

    return mapping


def extract_identifier(mapping, extra_keywords=None):
    """Return corresponding identifier from package *mapping*.

    :param mapping: package mapping. The package mapping must be in the form
        of::

            {
                "key": "foo",
                "package_name": "Foo",
                "installed_version": "1.11",
            }

    :param extra_keywords: List of :term:`extra requirement keywords
        <extras_require>` if required. Default is None.

    :return: Corresponding identifier (e.g. "Foo-1.11", "Bar").

    """
    if extra_keywords is not None and len(extra_keywords):
        extra_keywords = "-" + "-".join(extra_keywords)

    return wiz.filesystem.sanitize_value(
        "{name}{extra}-{version}".format(
            name=mapping["package_name"],
            version=mapping["installed_version"],
            extra=extra_keywords or ""
        )
    )


def extract_key(mapping, extra_keywords=None):
    """Compute key for package *mapping*.

    :param mapping: package mapping. The package mapping must be in the form
        of::

            {
                "key": "foo",
                "package_name": "Foo",
                "installed_version": "1.11",
            }

    :param extra_keywords: List of :term:`extra requirement keywords
        <extras_require>` if required. Default is None.

    :return: Corresponding key (e.g. "foo", "foo-test").

    """
    if extra_keywords is not None and len(extra_keywords):
        extra_keywords = "-" + "-".join(extra_keywords)

    return wiz.filesystem.sanitize_value(
        "{name}{extra}".format(
            name=mapping["key"],
            extra=extra_keywords or ""
        )
    )


def is_system_required(metadata):
    """Indicate whether package is platform-specific from *metadata*.

    Package `classifiers <https://pypi.org/classifiers/>`_ are retrieved from
    *metadata* to indicate if a specific operating system is required.

    :param metadata: string resulting from the `pip show -v
        <https://pip.pypa.io/en/stable/reference/pip_show/>`_ command.

    :return: Boolean value.

    """
    classifiers = re.findall("Operating System :: .*", metadata)

    # Check if the package is os independent.
    os_independent = (
        len(classifiers) == 1 and
        classifiers[0] == "Operating System :: OS Independent"
    )

    return len(classifiers) > 0 and not os_independent


def extract_command_mapping(metadata, extra_keywords=None):
    """Extract command mapping from entry points within *metadata*.

    Package :term:`Entry-Points` are retrieved from *metadata* to extract the
    corresponding commands. Each function defined as ``console_scripts`` will
    be used to create associated command.

    Provided *extra_keywords* are used when commands depend on optional
    dependencies.

    :param metadata: string resulting from the `pip show -v
        <https://pip.pypa.io/en/stable/reference/pip_show/>`_ command.

    :param extra_keywords: List of :term:`extra requirement keywords
        <extras_require>` if required. Default is None.

    :return: command mapping

        It should be in the form of::

            {
                "foo": "python -m foo",
                "bar": "python -m bar.test"
            }

    """
    mapping = {}

    # Convention: command name can only have alpha-numeric characters, hyphens
    # and points.
    entry_points = re.search(
        r"Entry-points:\n\s*\[console_scripts]\n((?:\s*.+\s*=\s*.+\s*\n)+)",
        metadata
    )
    if entry_points is not None:
        entry_points = entry_points.group(1)

        for element in entry_points.split("\n"):
            element = element.strip()
            if not len(element):
                continue

            alias, script = element.split("=")
            alias = alias.strip()
            script = script.strip()

            matched_extra = EXTRA_REQUEST_PATTERN.match(script)
            if matched_extra:
                authorized = set(matched_extra.group(1).split(","))
                if authorized.difference(extra_keywords or []):
                    continue

            command = script.split(":")[0].strip()
            if command.endswith(".__main__"):
                command = command[:-9]

            mapping[alias] = "python -m {}".format(command)

    return mapping


def extract_target_path(name, identifier, python_version, os_mapping=None):
    """Return the corresponding target path from package *mapping*.

    :param name: Python package name.

    :param identifier: Python package identifier.

    :param python_version: Python version identifier (e.g. "2.7").

    :param os_mapping: None or a mapping describing the operating system.
        Default is None.

        It should be in the form of::

            {
                "name": "centos",
                "major_version": 7
            }

    :return: Corresponding path (e.g. "Foo/Foo-1.11-py27-centos7").

    """
    path = os.path.join(name, identifier)

    # Indicate Python version
    path += "-py{}".format(python_version.replace(".", ""))

    # Indicate system if necessary
    if os_mapping:
        path += "-{os_name}{os_version}".format(
            os_name=os_mapping["name"],
            os_version=os_mapping["major_version"],
        )

    return path



# :coding: utf-8

import re

import pytest

import qip.command
import qip.system
import qip.environ
import qip.package


@pytest.fixture()
def mocked_system_query(mocker):
    """Return mocked command execute."""
    return mocker.patch.object(qip.system, "query")


@pytest.fixture()
def mocked_command_execute(mocker):
    """Return mocked command execute."""
    return mocker.patch.object(qip.command, "execute")


@pytest.fixture()
def mocked_fetch_mapping_from_environ(mocker):
    """Return mocked 'fetch_mapping_from_environ' function"""
    return mocker.patch.object(qip.package, "fetch_mapping_from_environ")


@pytest.fixture()
def mocked_extract_dependency_mapping(mocker):
    """Return mocked 'extract_dependency_mapping' function"""
    return mocker.patch.object(qip.package, "extract_dependency_mapping")


@pytest.fixture()
def mocked_extract_identifier(mocker):
    """Return mocked 'extract_identifier' function"""
    return mocker.patch.object(qip.package, "extract_identifier")


@pytest.fixture()
def mocked_extract_key(mocker):
    """Return mocked 'extract_key' function"""
    return mocker.patch.object(qip.package, "extract_key")


@pytest.fixture()
def mocked_is_system_required(mocker):
    """Return mocked 'is_system_required' function"""
    return mocker.patch.object(qip.package, "is_system_required")


@pytest.fixture()
def mocked_extract_command_mapping(mocker):
    """Return mocked 'extract_command_mapping' function"""
    return mocker.patch.object(qip.package, "extract_command_mapping")


@pytest.fixture()
def mocked_extract_target_path(mocker):
    """Return mocked 'extract_target_path' function"""
    return mocker.patch.object(qip.package, "extract_target_path")


@pytest.mark.parametrize("package, expected", [
    ("foo", "foo"),
    ("foo==0.1.0", "foo==0.1.0"),
    ("foo >= 7, < 8", "foo >= 7, < 8"),
    ("git@host:user/foo.git", "git+ssh://git@host/user/foo.git"),
    ("git@host1:bar/foo.git@0.1.0", "git+ssh://git@host1/bar/foo.git@0.1.0"),
    (
        "git@github.com:username/foo.git@dev",
        "git+ssh://git@github.com/username/foo.git@dev"
    ),
], ids=[
    "foo",
    "foo==0.1.0",
    "foo >= 7, < 8",
    "git@host:user/foo.git",
    "git@host1:bar/foo.git@0.1.0",
    "git@github.com:username/foo.git@dev"
])
def test_install(
    mocked_fetch_mapping_from_environ, mocked_command_execute, package, expected
):
    """Install package."""
    mocked_fetch_mapping_from_environ.return_value = {}
    mocked_command_execute.return_value = "Installing collected packages: foo"

    result = qip.package.install(
        package, "/path", {"environ": "__ENV__"}, "/cache"
    )
    mocked_fetch_mapping_from_environ.assert_called_once_with(
        "foo", {"environ": "__ENV__"}, extra_keywords=[]
    )
    assert result == {"request": expected}


def test_install_fail(mocked_command_execute):
    """Install package fails on pip install."""
    mocked_command_execute.return_value = "__FAIL__"

    with pytest.raises(ValueError) as error:
        qip.package.install("foo", "/path", {"environ": "__ENV__"}, "/cache")
    assert str(error.value) == "Package name could not be extracted from 'foo'."


def test_fetch_mapping_from_environ(
    mocked_extract_dependency_mapping, mocked_extract_identifier,
    mocked_extract_key, mocked_is_system_required,
    mocked_extract_command_mapping, mocked_extract_target_path,
    mocked_command_execute, mocked_system_query
):
    """Return a package mapping from environment."""
    context = {
        "environ": "__ENV__",
        "python": {
            "identifier": "2.8"
        }
    }

    package = {
        "key": "foo",
        "package_name": "Foo",
        "module_name": "foo",
        "installed_version": "0.1.0"
    }

    mocked_command_execute.return_value = ""
    mocked_extract_dependency_mapping.return_value = {
        "package": package,
        "requirements": []
    }

    mocked_extract_identifier.return_value = "Foo-0.1.0"
    mocked_extract_key.return_value = "foo"
    mocked_is_system_required.return_value = False
    mocked_extract_command_mapping.return_value = {}
    mocked_extract_target_path.return_value = "/path/to/target"

    mapping = qip.package.fetch_mapping_from_environ("foo", context)
    assert mapping == {
        "identifier": "Foo-0.1.0",
        "key": "foo",
        "name": "Foo",
        "module_name": "foo",
        "version": "0.1.0",
        "python": {
            "identifier": "2.8"
        },
        "target": "/path/to/target"
    }

    mocked_command_execute.assert_called_once_with(
        "python -m pip show --disable-pip-version-check 'foo' -v", "__ENV__",
        quiet=True
    )
    mocked_system_query.assert_not_called()
    mocked_extract_identifier.assert_called_once_with(
        package, extra_keywords=None
    )
    mocked_extract_key.assert_called_once_with(
        package, extra_keywords=None
    )


def test_fetch_mapping_from_environ_with_extra(
    mocked_extract_dependency_mapping, mocked_extract_identifier,
    mocked_extract_key, mocked_is_system_required,
    mocked_extract_command_mapping, mocked_extract_target_path,
    mocked_command_execute, mocked_system_query
):
    """Return a package mapping from environment with extra keywords."""
    context = {
        "environ": "__ENV__",
        "python": {
            "identifier": "2.8"
        }
    }

    package = {
        "key": "foo",
        "package_name": "Foo",
        "module_name": "foo",
        "installed_version": "0.1.0"
    }

    mocked_command_execute.return_value = ""
    mocked_extract_dependency_mapping.return_value = {
        "package": package,
        "requirements": []
    }

    mocked_extract_identifier.return_value = "Foo-doc-test-0.1.0"
    mocked_extract_key.return_value = "foo-doc-test"
    mocked_is_system_required.return_value = False
    mocked_extract_command_mapping.return_value = {}
    mocked_extract_target_path.return_value = "/path/to/target"

    mapping = qip.package.fetch_mapping_from_environ(
        "foo", context, extra_keywords=["doc", "test"]
    )
    assert mapping == {
        "identifier": "Foo-doc-test-0.1.0",
        "key": "foo-doc-test",
        "name": "Foo",
        "module_name": "foo",
        "version": "0.1.0",
        "python": {
            "identifier": "2.8"
        },
        "target": "/path/to/target"
    }

    mocked_command_execute.assert_called_once_with(
        "python -m pip show --disable-pip-version-check 'foo' -v", "__ENV__",
        quiet=True
    )
    mocked_system_query.assert_not_called()
    mocked_extract_identifier.assert_called_once_with(
        package, extra_keywords=["doc", "test"]
    )
    mocked_extract_key.assert_called_once_with(
        package, extra_keywords=["doc", "test"]
    )


def test_fetch_mapping_from_environ_with_system(
    mocked_extract_dependency_mapping, mocked_extract_identifier,
    mocked_extract_key, mocked_is_system_required,
    mocked_extract_command_mapping, mocked_extract_target_path,
    mocked_command_execute, mocked_system_query
):
    """Return a package mapping from environment with system."""
    context = {
        "environ": "__ENV__",
        "python": {
            "identifier": "2.8"
        }
    }

    package = {
        "key": "foo",
        "package_name": "Foo",
        "module_name": "foo",
        "installed_version": "0.1.0"
    }

    mocked_command_execute.return_value = ""
    mocked_extract_dependency_mapping.return_value = {
        "package": package,
        "requirements": []
    }

    mocked_extract_identifier.return_value = "Foo-0.1.0"
    mocked_extract_key.return_value = "foo"
    mocked_is_system_required.return_value = True
    mocked_extract_command_mapping.return_value = {}
    mocked_extract_target_path.return_value = "/path/to/target"
    mocked_system_query.return_value = {"os": "__SYSTEM__"}

    mapping = qip.package.fetch_mapping_from_environ("foo", context)
    assert mapping == {
        "identifier": "Foo-0.1.0",
        "key": "foo",
        "name": "Foo",
        "module_name": "foo",
        "version": "0.1.0",
        "python": {
            "identifier": "2.8"
        },
        "target": "/path/to/target",
        "system": {"os": "__SYSTEM__"}
    }

    mocked_command_execute.assert_called_once_with(
        "python -m pip show --disable-pip-version-check 'foo' -v", "__ENV__",
        quiet=True
    )
    mocked_system_query.assert_called_once()
    mocked_extract_identifier.assert_called_once_with(
        package, extra_keywords=None
    )
    mocked_extract_key.assert_called_once_with(
        package, extra_keywords=None
    )


def test_fetch_mapping_from_environ_with_description(
    mocked_extract_dependency_mapping, mocked_extract_identifier,
    mocked_extract_key, mocked_is_system_required,
    mocked_extract_command_mapping, mocked_extract_target_path,
    mocked_command_execute, mocked_system_query
):
    """Return a package mapping from environment with description."""
    context = {
        "environ": "__ENV__",
        "python": {
            "identifier": "2.8"
        }
    }

    package = {
        "key": "foo",
        "package_name": "Foo",
        "module_name": "foo",
        "installed_version": "0.1.0"
    }

    mocked_command_execute.return_value = "Summary: This is a test"
    mocked_extract_dependency_mapping.return_value = {
        "package": package,
        "requirements": []
    }

    mocked_extract_identifier.return_value = "Foo-0.1.0"
    mocked_extract_key.return_value = "foo"
    mocked_is_system_required.return_value = False
    mocked_extract_command_mapping.return_value = {}
    mocked_extract_target_path.return_value = "/path/to/target"

    mapping = qip.package.fetch_mapping_from_environ("foo", context)
    assert mapping == {
        "identifier": "Foo-0.1.0",
        "key": "foo",
        "name": "Foo",
        "module_name": "foo",
        "version": "0.1.0",
        "python": {
            "identifier": "2.8"
        },
        "target": "/path/to/target",
        "description": "This is a test"
    }

    mocked_command_execute.assert_called_once_with(
        "python -m pip show --disable-pip-version-check 'foo' -v", "__ENV__",
        quiet=True
    )
    mocked_system_query.assert_not_called()
    mocked_extract_identifier.assert_called_once_with(
        package, extra_keywords=None
    )
    mocked_extract_key.assert_called_once_with(
        package, extra_keywords=None
    )


def test_fetch_mapping_from_environ_with_location(
    mocked_extract_dependency_mapping, mocked_extract_identifier,
    mocked_extract_key, mocked_is_system_required,
    mocked_extract_command_mapping, mocked_extract_target_path,
    mocked_command_execute, mocked_system_query
):
    """Return a package mapping from environment with location."""
    context = {
        "environ": "__ENV__",
        "python": {
            "identifier": "2.8"
        }
    }

    package = {
        "key": "foo",
        "package_name": "Foo",
        "module_name": "foo",
        "installed_version": "0.1.0"
    }

    mocked_command_execute.return_value = "Location: /path/to/package"
    mocked_extract_dependency_mapping.return_value = {
        "package": package,
        "requirements": []
    }

    mocked_extract_identifier.return_value = "Foo-0.1.0"
    mocked_extract_key.return_value = "foo"
    mocked_is_system_required.return_value = False
    mocked_extract_command_mapping.return_value = {}
    mocked_extract_target_path.return_value = "/path/to/target"

    mapping = qip.package.fetch_mapping_from_environ("foo", context)
    assert mapping == {
        "identifier": "Foo-0.1.0",
        "key": "foo",
        "name": "Foo",
        "module_name": "foo",
        "version": "0.1.0",
        "python": {
            "identifier": "2.8"
        },
        "target": "/path/to/target",
        "location": "/path/to/package"
    }

    mocked_command_execute.assert_called_once_with(
        "python -m pip show --disable-pip-version-check 'foo' -v", "__ENV__",
        quiet=True
    )
    mocked_system_query.assert_not_called()
    mocked_extract_identifier.assert_called_once_with(
        package, extra_keywords=None
    )
    mocked_extract_key.assert_called_once_with(
        package, extra_keywords=None
    )


def test_fetch_mapping_from_environ_with_commands(
    mocked_extract_dependency_mapping, mocked_extract_identifier,
    mocked_extract_key, mocked_is_system_required,
    mocked_extract_command_mapping, mocked_extract_target_path,
    mocked_command_execute, mocked_system_query
):
    """Return a package mapping from environment with commands."""
    context = {
        "environ": "__ENV__",
        "python": {
            "identifier": "2.8"
        }
    }

    package = {
        "key": "foo",
        "package_name": "Foo",
        "module_name": "foo",
        "installed_version": "0.1.0"
    }

    mocked_command_execute.return_value = ""
    mocked_extract_dependency_mapping.return_value = {
        "package": package,
        "requirements": []
    }

    mocked_extract_identifier.return_value = "Foo-0.1.0"
    mocked_extract_key.return_value = "foo"
    mocked_is_system_required.return_value = False
    mocked_extract_command_mapping.return_value = {"foo": "python -m foo"}
    mocked_extract_target_path.return_value = "/path/to/target"

    mapping = qip.package.fetch_mapping_from_environ("foo", context)
    assert mapping == {
        "identifier": "Foo-0.1.0",
        "key": "foo",
        "name": "Foo",
        "module_name": "foo",
        "version": "0.1.0",
        "python": {
            "identifier": "2.8"
        },
        "target": "/path/to/target",
        "command": {"foo": "python -m foo"}
    }

    mocked_command_execute.assert_called_once_with(
        "python -m pip show --disable-pip-version-check 'foo' -v", "__ENV__",
        quiet=True
    )
    mocked_system_query.assert_not_called()
    mocked_extract_identifier.assert_called_once_with(
        package, extra_keywords=None
    )
    mocked_extract_key.assert_called_once_with(
        package, extra_keywords=None
    )


def test_fetch_mapping_from_environ_with_requirements(
    mocked_extract_dependency_mapping, mocked_extract_identifier,
    mocked_extract_key, mocked_is_system_required,
    mocked_extract_command_mapping, mocked_extract_target_path,
    mocked_command_execute, mocked_system_query
):
    """Return a package mapping from environment with requirements."""
    context = {
        "environ": "__ENV__",
        "python": {
            "identifier": "2.8"
        }
    }

    package = {
        "key": "foo",
        "package_name": "Foo",
        "module_name": "foo",
        "installed_version": "0.1.0"
    }

    mocked_command_execute.return_value = ""
    mocked_extract_dependency_mapping.return_value = {
        "package": package,
        "requirements": [
            "bim >= 0.1.0, < 1",
            "baz",
        ]
    }

    mocked_extract_identifier.return_value = "Foo-0.1.0"
    mocked_extract_key.return_value = "foo"
    mocked_is_system_required.return_value = False
    mocked_extract_command_mapping.return_value = {}
    mocked_extract_target_path.return_value = "/path/to/target"

    mapping = qip.package.fetch_mapping_from_environ("foo", context)
    assert mapping == {
        "identifier": "Foo-0.1.0",
        "key": "foo",
        "name": "Foo",
        "module_name": "foo",
        "version": "0.1.0",
        "python": {
            "identifier": "2.8"
        },
        "target": "/path/to/target",
        "requirements": [
            "bim >= 0.1.0, < 1",
            "baz"
        ]
    }

    mocked_command_execute.assert_called_once_with(
        "python -m pip show --disable-pip-version-check 'foo' -v", "__ENV__",
        quiet=True
    )
    mocked_system_query.assert_not_called()
    mocked_extract_identifier.assert_called_once_with(
        package, extra_keywords=None
    )
    mocked_extract_key.assert_called_once_with(
        package, extra_keywords=None
    )


def test_extract_dependency_mapping(mocker, mocked_command_execute):
    """Return package mapping from dependency mapping."""
    mocked_command_execute.return_value = "{\"package\": {\"key\": \"foo\"}}"

    mapping = qip.package.extract_dependency_mapping("foo", {})

    mocked_command_execute.assert_called_once_with(mocker.ANY, {}, quiet=True)
    assert re.match(
        "python .+ foo", mocked_command_execute.call_args_list[0][0][0]
    )
    assert mapping == {"package": {"key": "foo"}}


def test_extract_dependency_mapping_with_extra(mocker, mocked_command_execute):
    """Return package mapping from dependency mapping with extra requirements.
    """
    mocked_command_execute.return_value = "{\"package\": {\"key\": \"foo\"}}"

    mapping = qip.package.extract_dependency_mapping(
        "foo", {}, extra_keywords=["doc", "test"]
    )
    mocked_command_execute.assert_called_once_with(mocker.ANY, {}, quiet=True)
    assert re.match(
        r"python .+ foo\[doc,test]",
        mocked_command_execute.call_args_list[0][0][0]
    )
    assert mapping == {"package": {"key": "foo"}}


def test_extract_dependency_mapping_fail_tree(mocked_command_execute):
    """Fail to return package mapping if pip_query returns nothing."""
    mocked_command_execute.return_value = ""

    with pytest.raises(RuntimeError) as error:
        qip.package.extract_dependency_mapping("foo", {})

    assert str(error.value) == "Impossible to fetch installed package for 'foo'"


def test_extract_identifier():
    """Return identifier from package mapping."""
    mapping = {
        "key": "foo",
        "package_name": "Foo",
        "module_name": "foo",
        "installed_version": "1.11",
    }
    identifier = qip.package.extract_identifier(mapping)
    assert identifier == "Foo-1.11"


def test_extract_identifier_with_extra():
    """Return identifier computed with extra keywords."""
    mapping = {
        "key": "foo",
        "package_name": "Foo",
        "module_name": "foo",
        "installed_version": "1.11",
    }
    identifier = qip.package.extract_identifier(
        mapping, extra_keywords=["doc", "test"]
    )
    assert identifier == "Foo-doc-test-1.11"


def test_extract_key():
    """Return key from package mapping."""
    mapping = {
        "key": "foo",
        "package_name": "Foo",
        "module_name": "foo",
        "installed_version": "1.11",
    }
    identifier = qip.package.extract_key(mapping)
    assert identifier == "foo"


def test_extract_key_with_extra():
    """Return key from package mapping with extra keywords."""
    mapping = {
        "key": "foo",
        "package_name": "Foo",
        "module_name": "foo",
        "installed_version": "1.11",
    }
    identifier = qip.package.extract_key(
        mapping, extra_keywords=["doc", "test"]
    )
    assert identifier == "foo-doc-test"


@pytest.mark.parametrize("metadata, expected", [
    ("", False),
    ("Operating System :: OS Independent", False),
    (
        (
            "Operating System :: OS Independent"
            "Operating System :: Linux"
        ),
        True
    ),
    (
        (
            "Operating System :: Mac"
            "Operating System :: Linux"
        ),
        True
    ),
    (
        (
            "Operating System :: Mac"
        ),
        True
    ),
], ids=[
    "no-metadata",
    "independent",
    "confusing-classifiers",
    "multi-platforms",
    "single-platform",
])
def test_is_system_required(metadata, expected):
    """Indicate whether package is platform-specific."""
    assert qip.package.is_system_required(metadata) == expected


@pytest.mark.parametrize("metadata, expected", [
    ("", {}),
    (
        (
            "Entry-points:\n"
            "  [console_scripts]\n"
            "  sphinx-apidoc = sphinx.ext.apidoc:main\n"
            "  sphinx-autogen = sphinx.ext.autosummary.generate:main\n"
            "  sphinx-build = sphinx.cmd.build:main\n"
            "  sphinx-quickstart = sphinx.cmd.quickstart:main\n"
        ),
        {
            "sphinx-apidoc": "python -m sphinx.ext.apidoc",
            "sphinx-autogen": "python -m sphinx.ext.autosummary.generate",
            "sphinx-build": "python -m sphinx.cmd.build",
            "sphinx-quickstart": "python -m sphinx.cmd.quickstart",
        }
    ),
    (
        (
            "Entry-points:\n"
            "  [console_scripts]\n"
            "  qip = qip.__main__:main\n"
        ),
        {
            "qip": "python -m qip"
        }
    ),
], ids=[
    "no-metadata",
    "multi-commands",
    "main-command",
])
def test_extract_command_mapping(metadata, expected):
    """Extract command mapping from entry points"""
    assert qip.package.extract_command_mapping(metadata) == expected


def test_extract_command_mapping_with_extra():
    """Extract command mapping from entry points with extra keywords"""
    metadata = (
        "Entry-points:\n"
        "  [console_scripts]\n"
        "  foo = foo.__main__:main [test]\n"
        "  bar = bar.__main__:main [doc]\n"
    )

    assert qip.package.extract_command_mapping(metadata) == {}

    assert qip.package.extract_command_mapping(
        metadata, extra_keywords=["test"]
    ) == {
        "foo": "python -m foo"
    }

    assert qip.package.extract_command_mapping(
        metadata, extra_keywords=["doc"]
    ) == {
        "bar": "python -m bar"
    }

    assert qip.package.extract_command_mapping(
        metadata, extra_keywords=["doc", "test"]
    ) == {
        "foo": "python -m foo",
        "bar": "python -m bar"
    }


def test_extract_target_path():
    """Return the corresponding target path from package."""
    path = qip.package.extract_target_path("Foo", "Foo-0.1.0", "2.8")
    assert path == "Foo/Foo-0.1.0-py28"


def test_extract_target_path_with_system():
    """Return the corresponding target path from package with system."""
    path = qip.package.extract_target_path(
        "Foo", "Foo-0.1.0", "2.8", os_mapping={
            "name": "centos",
            "major_version": 7
        }
    )
    assert path == "Foo/Foo-0.1.0-py28-centos7"



# :coding: utf-8

import platform

import distro
import pytest
from packaging.version import InvalidVersion

import qip.system


@pytest.fixture()
def mocked_platform_system(mocker):
    """Mocked the platform.system function."""
    return mocker.patch.object(platform, "system")


@pytest.fixture()
def mocked_platform_machine(mocker):
    """Mocked the platform.machine function."""
    return mocker.patch.object(platform, "machine")


@pytest.fixture()
def mocked_platform_linux(mocker):
    """Mocked the distro.linux_distribution function."""
    return mocker.patch.object(distro, "linux_distribution")


@pytest.fixture()
def mocked_platform_mac(mocker):
    """Mocked the platform.mac_ver function."""
    return mocker.patch.object(platform, "mac_ver")


@pytest.fixture()
def mocked_platform_win32(mocker):
    """Mocked the platform.win32_ver function."""
    return mocker.patch.object(platform, "win32_ver")


@pytest.fixture()
def mocked_query_linux(mocker):
    """Mocked the query_linux function."""
    return mocker.patch.object(
        qip.system, "query_linux", return_value="LINUX_SYSTEM_MAPPING"
    )


@pytest.fixture()
def mocked_query_mac(mocker):
    """Mocked the query_mac function."""
    return mocker.patch.object(
        qip.system, "query_mac", return_value="MAC_SYSTEM_MAPPING"
    )


@pytest.fixture()
def mocked_query_windows(mocker):
    """Mocked the query_windows function."""
    return mocker.patch.object(
        qip.system, "query_windows", return_value="WINDOWS_SYSTEM_MAPPING"
    )


@pytest.mark.parametrize("platform, expected", [
    ("Linux", "LINUX_SYSTEM_MAPPING"),
    ("Darwin", "MAC_SYSTEM_MAPPING"),
    ("Windows", "WINDOWS_SYSTEM_MAPPING")
], ids=[
    "linux",
    "mac",
    "windows",
])
@pytest.mark.usefixtures("mocked_query_linux")
@pytest.mark.usefixtures("mocked_query_mac")
@pytest.mark.usefixtures("mocked_query_windows")
def test_query(mocked_platform_system, platform, expected):
    """Query system mapping."""
    mocked_platform_system.return_value = platform
    assert qip.system.query() == expected


def test_query_version_error(mocked_platform_system, mocked_platform_linux):
    """Fails to query system mapping from incorrect os version."""
    mocked_platform_system.return_value = "linux"
    mocked_platform_linux.side_effect = InvalidVersion

    with pytest.raises(RuntimeError):
        qip.system.query()


@pytest.mark.parametrize("distribution, architecture, expected", [
    (
        ("centos", "7.3.1611", "Core"), "x86_64",
        {
            "platform": "linux",
            "arch": "x86_64",
            "os": {"name": "centos", "major_version": 7}
        }
    ),
    (
        ("centos", "6.5", "Final"), "x86_64",
        {
            "platform": "linux",
            "arch": "x86_64",
            "os": {"name": "centos", "major_version": 6}
        }
    ),
    (
        ("redhat", "5.7", "Final"), "i386",
        {
            "platform": "linux",
            "arch": "i386",
            "os": {"name": "redhat", "major_version": 5}
        }
    )
], ids=[
    "centos-7-64b",
    "centos-6-64b",
    "redhat-5-32b",
])
def test_query_linux(
    mocked_platform_linux, mocked_platform_machine,
    distribution, architecture, expected
):
    """Query linux system mapping."""
    mocked_platform_linux.return_value = distribution
    mocked_platform_machine.return_value = architecture
    assert qip.system.query_linux() == expected


@pytest.mark.parametrize("mac_ver, architecture, expected", [
    (
        ("10.13.3", ("", "", ""), ""), "x86_64",
        {
            "platform": "mac",
            "arch": "x86_64",
            "os": {"name": "mac", "major_version": 10}
        }
    ),
    (
        ("9.0.0", ("", "", ""), ""), "x86_64",
        {
            "platform": "mac",
            "arch": "x86_64",
            "os": {"name": "mac", "major_version": 9}
        }
    )
], ids=[
    "mac-10.13",
    "mac-9.0",
])
def test_query_mac(
    mocked_platform_mac, mocked_platform_machine,
    mac_ver, architecture, expected
):
    """Query mac system mapping."""
    mocked_platform_mac.return_value = mac_ver
    mocked_platform_machine.return_value = architecture
    assert qip.system.query_mac() == expected


@pytest.mark.parametrize("win32_ver, architecture, expected", [
    (
        ("XP", "5.1.2600", "SP2", "Multiprocessor Free"), "i386",
        {
            "platform": "windows",
            "arch": "i386",
            "os": {"name": "windows", "major_version": 5}
        }
    ),
    (
        ("10", "10.0.16299", "", "Multiprocessor Free"), "AMD64",
        {
            "platform": "windows",
            "arch": "AMD64",
            "os": {"name": "windows", "major_version": 10}
        }
    )
], ids=[
    "windows-xp",
    "windows-10",
])
def test_query_windows(
    mocked_platform_win32, mocked_platform_machine,
    win32_ver, architecture, expected
):
    """Query mac system mapping."""
    mocked_platform_win32.return_value = win32_ver
    mocked_platform_machine.return_value = architecture
    assert qip.system.query_windows() == expected


def test_extract_major_version():
    """Extract major version of operating system request for *version*."""
    major_version = qip.system._extract_major_version("10.5.0")
    assert major_version == 10



# :coding: utf-8

import os
import functools

import pytest
import wiz
import wiz.exception
import wiz.definition

import qip.definition


@pytest.fixture()
def mocked_wiz_export_definition(mocker):
    """Return mocked 'wiz.export_definition' function"""
    return mocker.patch.object(wiz, "export_definition")


@pytest.fixture()
def mocked_wiz_load_definition(mocker):
    """Return mocked 'wiz.load_definition' function"""
    return mocker.patch.object(wiz, "load_definition")


@pytest.fixture()
def mocked_wiz_fetch_definition(mocker):
    """Return mocked 'wiz.fetch_definition' function"""
    return mocker.patch.object(wiz, "fetch_definition")


@pytest.fixture()
def mocked_update(mocker):
    """Return mocked 'qip.definition.update' function"""
    return mocker.patch.object(qip.definition, "update")


@pytest.fixture()
def mocked_create(mocker):
    """Return mocked 'qip.definition.create' function"""
    return mocker.patch.object(qip.definition, "create")


@pytest.mark.parametrize("options, additional_variants, editable_mode", [
    ({}, None, False),
    ({"editable_mode": True}, None, True),
    (
        {
            "existing_definition": wiz.definition.Definition({
                "identifier": "foo",
                "variants": [{"identifier": "V3", "environ": {"KEY": "VALUE"}}]
            })
        },
        [{"identifier": "V3", "environ": {"KEY": "VALUE"}}], False
    ),
], ids=[
    "simple",
    "with-editable-mode",
    "with-additional-variants",
])
def test_export(
    mocked_update, mocked_create, mocked_wiz_export_definition, options,
    additional_variants, editable_mode
):
    """Export definition."""
    mapping = {"request": "foo >= 1, < 2"}
    path = "/definitions"
    output_path = "/packages"

    qip.definition.export(path, mapping, output_path, **options)

    mocked_update.assert_not_called()
    mocked_create.assert_called_once_with(
        mapping, output_path, editable_mode=editable_mode,
        additional_variants=additional_variants
    )
    mocked_wiz_export_definition.assert_called_once_with(
        path, mocked_create.return_value, overwrite=True
    )


@pytest.mark.parametrize("options, additional_variants, editable_mode", [
    ({}, None, False),
    ({"editable_mode": True}, None, True),
    (
        {
            "existing_definition": wiz.definition.Definition({
                "identifier": "foo",
                "variants": [{"identifier": "V3", "environ": {"KEY": "VALUE"}}]
            })
        },
        [{"identifier": "V3", "environ": {"KEY": "VALUE"}}], False
    ),
], ids=[
    "simple",
    "with-editable-mode",
    "with-additional-variants",
])
def test_export_from_custom_definition(
    mocked_update, mocked_create, mocked_wiz_export_definition, options,
    additional_variants, editable_mode
):
    """Export definition from custom definition retrieved."""
    mapping = {"request": "foo >= 1, < 2"}
    path = "/definitions"
    output_path = "/packages"

    qip.definition.export(
        path, mapping, output_path,
        custom_definition="__CUSTOM_DEFINITION__",
        **options
    )

    mocked_update.assert_called_once_with(
        "__CUSTOM_DEFINITION__", mapping, output_path,
        editable_mode=editable_mode,
        additional_variants=additional_variants
    )
    mocked_create.assert_not_called()
    mocked_wiz_export_definition.assert_called_once_with(
        path, mocked_update.return_value, overwrite=True
    )


def test_fetch_custom(mocked_wiz_load_definition, temporary_directory, logger):
    """Fetch custom definition from package installed."""
    mapping = {
        "identifier": "Foo-0.2.3",
        "module_name": "foo",
        "location": temporary_directory
    }

    path = os.path.join(temporary_directory, "foo", "package_data", "wiz.json")

    os.makedirs(os.path.dirname(path))
    with open(path, "w") as stream:
        stream.write("")

    definition = wiz.definition.Definition({
        "identifier": "foo",
        "version": "0.2.3",
        "namespace": "plugin",
        "description": "This is a library",
    })

    mocked_wiz_load_definition.return_value = definition

    _definition = qip.definition.fetch_custom(mapping)

    assert _definition.data() == {
        "identifier": "foo",
        "version": "0.2.3",
        "namespace": "plugin",
        "description": "This is a library",
    }

    logger.info.assert_called_once_with(
        "\tWiz definition extracted from 'Foo-0.2.3'."
    )

    mocked_wiz_load_definition.assert_any_call(path)


def test_fetch_custom_empty(mocked_wiz_load_definition, logger):
    """Fail to fetch custom definition from package mapping."""
    mapping = {
        "identifier": "Foo-0.2.3",
        "module_name": "foo",
        "location": "/path/to/lib"
    }

    result = qip.definition.fetch_custom(mapping)
    assert result is None

    mocked_wiz_load_definition.assert_not_called()
    logger.info.assert_not_called()


@pytest.mark.parametrize("options, namespace", [
    ({}, "library"),
    ({"namespace": "test"}, "test"),
], ids=[
    "simple",
    "with-namespace",
])
def test_fetch_existing(mocked_wiz_fetch_definition, options, namespace):
    """Fetch existing definition in definition mapping."""
    result = qip.definition.fetch_existing(
        {"key": "foo", "version": "0.1.0"}, "__MAPPING__", **options
    )
    mocked_wiz_fetch_definition.assert_called_once_with(
        "{0}::foo==0.1.0".format(namespace), "__MAPPING__"
    )
    assert result == mocked_wiz_fetch_definition.return_value


@pytest.mark.parametrize("options, namespace", [
    ({}, "library"),
    ({"namespace": "test"}, "test"),
], ids=[
    "simple",
    "with-namespace",
])
def test_fetch_existing_empty(mocked_wiz_fetch_definition, options, namespace):
    """Fail to fetch existing definition in definition mapping."""
    mocked_wiz_fetch_definition.side_effect = (
        wiz.exception.RequestNotFound("Error!")
    )

    result = qip.definition.fetch_existing(
        {"key": "foo", "version": "0.1.0"}, "__MAPPING__", **options
    )
    mocked_wiz_fetch_definition.assert_called_once_with(
        "{0}::foo==0.1.0".format(namespace), "__MAPPING__"
    )
    assert result is None


def test_create(logger):
    """Create a definition from package mapping."""
    mapping = {
        "identifier": "Foo-0.2.3",
        "name": "Foo",
        "key": "foo",
        "version": "0.2.3",
        "target": "Foo/Foo-0.2.3",
        "description": "This is a package",
        "python": {
            "identifier": "2.8",
            "request": "python >= 2.8, <2.9",
            "library-path": "lib/python2.8/site-packages"
        }
    }

    result = qip.definition.create(
        mapping, "/path/to/installed/package", editable_mode=False
    )
    assert result.data() == {
        "identifier": "foo",
        "version": "0.2.3",
        "install-root": "/path/to/installed/package",
        "namespace": "library",
        "description": "This is a package",
        "environ": {
            "PYTHONPATH": "${INSTALL_LOCATION}:${PYTHONPATH}"
        },
        "variants": [
            {
                "identifier": "2.8",
                "install-location": (
                    "${INSTALL_ROOT}/Foo/Foo-0.2.3/lib/python2.8/site-packages"
                ),
                "requirements": [
                    "python >= 2.8, <2.9",
                ]
            }
        ]
    }

    logger.info.assert_called_once_with(
        "\tWiz definition created for 'Foo-0.2.3'."
    )


def test_create_with_system(logger):
    """Create a definition from package mapping with system."""
    mapping = {
        "identifier": "Foo-0.2.3",
        "name": "Foo",
        "key": "foo",
        "version": "0.2.3",
        "system": {
            "platform": "linux",
            "arch": "x86_64",
            "os": {
                "name": "centos",
                "major_version": 7
            }
        },
        "target": "Foo/Foo-0.2.3",
        "description": "This is a package",
        "python": {
            "identifier": "2.8",
            "request": "python >= 2.8, <2.9",
            "library-path": "lib/python2.8/site-packages"
        }
    }

    result = qip.definition.create(
        mapping, "/path/to/installed/package", editable_mode=False
    )
    assert result.data() == {
        "identifier": "foo",
        "version": "0.2.3",
        "install-root": "/path/to/installed/package",
        "namespace": "library",
        "description": "This is a package",
        "environ": {
            "PYTHONPATH": "${INSTALL_LOCATION}:${PYTHONPATH}"
        },
        "system": {
            "platform": "linux",
            "arch": "x86_64",
            "os": "centos >= 7, < 8"
        },
        "variants": [
            {
                "identifier": "2.8",
                "install-location": (
                    "${INSTALL_ROOT}/Foo/Foo-0.2.3/lib/python2.8/site-packages"
                ),
                "requirements": [
                    "python >= 2.8, <2.9",
                ]
            }
        ]
    }

    logger.info.assert_called_once_with(
        "\tWiz definition created for 'Foo-0.2.3'."
    )


def test_create_with_requirements(logger):
    """Create a definition from package mapping with requirements."""
    mapping = {
        "identifier": "Foo-0.2.3",
        "name": "Foo",
        "key": "foo",
        "version": "0.2.3",
        "requirements": [
            "bim >= 3, < 4",
            "bar",
        ],
        "target": "Foo/Foo-0.2.3",
        "description": "This is a package",
        "python": {
            "identifier": "2.8",
            "request": "python >= 2.8, <2.9",
            "library-path": "lib/python2.8/site-packages"
        }
    }

    result = qip.definition.create(
        mapping, "/path/to/installed/package", editable_mode=False
    )
    assert result.data() == {
        "identifier": "foo",
        "version": "0.2.3",
        "install-root": "/path/to/installed/package",
        "namespace": "library",
        "description": "This is a package",
        "environ": {
            "PYTHONPATH": "${INSTALL_LOCATION}:${PYTHONPATH}"
        },
        "variants": [
            {
                "identifier": "2.8",
                "install-location": (
                    "${INSTALL_ROOT}/Foo/Foo-0.2.3/lib/python2.8/site-packages"
                ),
                "requirements": [
                    "python >= 2.8, <2.9",
                    "library::bim[2.8] >=3, <4",
                    "library::bar[2.8]",
                ]
            }
        ]
    }

    logger.info.assert_called_once_with(
        "\tWiz definition created for 'Foo-0.2.3'."
    )


def test_create_with_commands(logger):
    """Create a definition from package mapping with commands."""
    mapping = {
        "identifier": "Foo-0.2.3",
        "name": "Foo",
        "key": "foo",
        "version": "0.2.3",
        "target": "Foo/Foo-0.2.3",
        "description": "This is a package",
        "command": {
            "foo": "python -m foo",
            "test": "python -m foo.test",
        },
        "python": {
            "identifier": "2.8",
            "request": "python >= 2.8, <2.9",
            "library-path": "lib/python2.8/site-packages"
        }
    }

    result = qip.definition.create(
        mapping, "/path/to/installed/package", editable_mode=False
    )
    assert result.data() == {
        "identifier": "foo",
        "version": "0.2.3",
        "install-root": "/path/to/installed/package",
        "namespace": "library",
        "description": "This is a package",
        "command": {
            "foo": "python -m foo",
            "test": "python -m foo.test",
        },
        "environ": {
            "PYTHONPATH": "${INSTALL_LOCATION}:${PYTHONPATH}"
        },
        "variants": [
            {
                "identifier": "2.8",
                "install-location": (
                    "${INSTALL_ROOT}/Foo/Foo-0.2.3/lib/python2.8/site-packages"
                ),
                "requirements": [
                    "python >= 2.8, <2.9"
                ]
            }
        ]
    }

    logger.info.assert_called_once_with(
        "\tWiz definition created for 'Foo-0.2.3'."
    )


def test_create_editable_mode(logger):
    """Create a definition from package mapping in editable mode."""
    mapping = {
        "identifier": "Foo-0.2.3",
        "name": "Foo",
        "key": "foo",
        "version": "0.2.3",
        "target": "Foo/Foo-0.2.3",
        "description": "This is a package",
        "python": {
            "identifier": "2.8",
            "request": "python >= 2.8, <2.9",
            "library-path": "lib/python2.8/site-packages"
        },
        "location": "/path/to/source"
    }

    result = qip.definition.create(
        mapping, "/path/to/installed/package", editable_mode=True
    )
    assert result.data() == {
        "identifier": "foo",
        "version": "0.2.3",
        "namespace": "library",
        "description": "This is a package",
        "environ": {
            "PYTHONPATH": "${INSTALL_LOCATION}:${PYTHONPATH}"
        },
        "variants": [
            {
                "identifier": "2.8",
                "install-location": "/path/to/source",
                "requirements": [
                    "python >= 2.8, <2.9"
                ]
            }
        ]
    }

    logger.info.assert_called_once_with(
        "\tWiz definition created for 'Foo-0.2.3'."
    )


def test_create_with_additional_variants_1(logger):
    """Create a definition from package mapping with additional variants."""
    mapping = {
        "identifier": "Foo-0.2.3",
        "name": "Foo",
        "key": "foo",
        "version": "0.2.3",
        "requirements": [
            "bim >= 3, < 4",
            "bar",
        ],
        "target": "Foo/Foo-0.2.3",
        "description": "This is a package",
        "python": {
            "identifier": "2.8",
            "request": "python >= 2.8, <2.9",
            "library-path": "lib/python2.8/site-packages"
        }
    }

    result = qip.definition.create(
        mapping, "/path/to/installed/package",
        editable_mode=False,
        additional_variants=[
            {
                "identifier": "variant",
            },
            {
                "identifier": "3.6",
                "environ": {"KEY36": "VALUE36"}
            },
            {
                "identifier": "2.8",
                "environ": {"KEY28": "VALUE28"},
                "requirements": [
                    "library::bim[2.8] >= 3, < 4",
                    "library::baz"
                ]
            },
            {
                "identifier": "2.7",
                "environ": {"KEY22": "VALUE22"}
            }
        ]
    )
    assert result.data() == {
        "identifier": "foo",
        "version": "0.2.3",
        "install-root": "/path/to/installed/package",
        "namespace": "library",
        "description": "This is a package",
        "environ": {
            "PYTHONPATH": "${INSTALL_LOCATION}:${PYTHONPATH}",
        },
        "variants": [
            {
                "identifier": "3.6",
                "environ": {"KEY36": "VALUE36"}
            },
            {
                "identifier": "2.8",
                "install-location": (
                    "${INSTALL_ROOT}/Foo/Foo-0.2.3/lib/python2.8/site-packages"
                ),
                "environ": {
                    "KEY28": "VALUE28"
                },
                "requirements": [
                    "library::bim[2.8] >= 3, < 4",
                    "library::baz",
                    "python >= 2.8, <2.9",
                    "library::bar[2.8]",
                ]
            },
            {
                "identifier": "2.7",
                "environ": {"KEY22": "VALUE22"}
            },
            {
                "identifier": "variant",
            }
        ]
    }

    logger.info.assert_called_once_with(
        "\tWiz definition created for 'Foo-0.2.3'."
    )


def test_create_with_additional_variants_2(logger):
    """Create a definition from package mapping with additional variants."""
    mapping = {
        "identifier": "Foo-0.2.3",
        "name": "Foo",
        "key": "foo",
        "version": "0.2.3",
        "requirements": [
            "bim >= 3, < 4",
            "bar",
        ],
        "target": "Foo/Foo-0.2.3",
        "description": "This is a package",
        "python": {
            "identifier": "2.8",
            "request": "python >= 2.8, <2.9",
            "library-path": "lib/python2.8/site-packages"
        }
    }

    result = qip.definition.create(
        mapping, "/path/to/installed/package",
        editable_mode=False,
        additional_variants=[
            {
                "identifier": "variant",
            },
            {
                "identifier": "3.6",
                "environ": {"KEY36": "VALUE36"}
            },
            {
                "identifier": "2.7",
                "environ": {"KEY22": "VALUE22"}
            }
        ]
    )
    assert result.data() == {
        "identifier": "foo",
        "version": "0.2.3",
        "install-root": "/path/to/installed/package",
        "namespace": "library",
        "description": "This is a package",
        "environ": {
            "PYTHONPATH": "${INSTALL_LOCATION}:${PYTHONPATH}",
        },
        "variants": [
            {
                "identifier": "3.6",
                "environ": {"KEY36": "VALUE36"}
            },
            {
                "identifier": "2.8",
                "install-location": (
                    "${INSTALL_ROOT}/Foo/Foo-0.2.3/lib/python2.8/site-packages"
                ),
                "requirements": [
                    "python >= 2.8, <2.9",
                    "library::bim[2.8] >=3, <4",
                    "library::bar[2.8]",
                ]
            },
            {
                "identifier": "2.7",
                "environ": {"KEY22": "VALUE22"}
            },
            {
                "identifier": "variant",
            }
        ]
    }

    logger.info.assert_called_once_with(
        "\tWiz definition created for 'Foo-0.2.3'."
    )


def test_update():
    """Update definition with mapping."""
    definition = wiz.definition.Definition({
        "identifier": "foo",
        "version": "0.2.3",
        "namespace": "plugin",
        "description": "This is a library",
    })

    mapping = {
        "identifier": "Foo-0.2.3",
        "name": "Foo",
        "key": "foo",
        "version": "0.2.3",
        "description": "This is a cool library",
        "target": "Foo/Foo-0.2.3",
        "python": {
            "identifier": "2.8",
            "request": "python >= 2.8, <2.9",
            "library-path": "lib/python2.8/site-packages"
        },
    }

    result = qip.definition.update(definition, mapping, "/packages")

    assert result.data() == {
        "identifier": "foo",
        "version": "0.2.3",
        "namespace": "plugin",
        "description": "This is a library",
        "install-root": "/packages",
        "environ": {
            "PYTHONPATH": "${INSTALL_LOCATION}:${PYTHONPATH}"
        },
        "variants": [
            {
                "identifier": "2.8",
                "install-location": (
                    "${INSTALL_ROOT}/Foo/Foo-0.2.3/lib/python2.8/site-packages"
                ),
                "requirements": [
                    "python >= 2.8, <2.9"
                ]
            }
        ]
    }


def test_update_with_pythonpath():
    """Update definition with existing PYTHONPATH from mapping."""
    definition = wiz.definition.Definition({
        "identifier": "foo",
        "version": "0.2.3",
        "namespace": "plugin",
        "description": "This is a library",
        "environ": {
            "PYTHONPATH": "/path/to/somewhere:${PYTHONPATH}"
        }
    })

    mapping = {
        "identifier": "Foo-0.2.3",
        "name": "Foo",
        "key": "foo",
        "version": "0.2.3",
        "description": "This is a cool library",
        "target": "Foo/Foo-0.2.3",
        "python": {
            "identifier": "2.8",
            "request": "python >= 2.8, <2.9",
            "library-path": "lib/python2.8/site-packages"
        },
    }

    result = qip.definition.update(definition, mapping, "/packages")

    assert result.data() == {
        "identifier": "foo",
        "version": "0.2.3",
        "namespace": "plugin",
        "description": "This is a library",
        "install-root": "/packages",
        "environ": {
            "PYTHONPATH": "${INSTALL_LOCATION}:/path/to/somewhere:${PYTHONPATH}"
        },
        "variants": [
            {
                "identifier": "2.8",
                "install-location": (
                    "${INSTALL_ROOT}/Foo/Foo-0.2.3/lib/python2.8/site-packages"
                ),
                "requirements": [
                    "python >= 2.8, <2.9"
                ]
            }
        ]
    }


def test_update_without_description():
    """Update definition without description."""
    definition = wiz.definition.Definition({
        "identifier": "foo",
        "version": "0.2.3",
        "namespace": "plugin",
    })

    mapping = {
        "identifier": "Foo-0.2.3",
        "name": "Foo",
        "key": "foo",
        "version": "0.2.3",
        "description": "This is a cool library",
        "target": "Foo/Foo-0.2.3",
        "python": {
            "identifier": "2.8",
            "request": "python >= 2.8, <2.9",
            "library-path": "lib/python2.8/site-packages"
        },
    }

    result = qip.definition.update(definition, mapping, "/packages")

    assert result.data() == {
        "identifier": "foo",
        "version": "0.2.3",
        "namespace": "plugin",
        "description": "This is a cool library",
        "install-root": "/packages",
        "environ": {
            "PYTHONPATH": "${INSTALL_LOCATION}:${PYTHONPATH}"
        },
        "variants": [
            {
                "identifier": "2.8",
                "install-location": (
                    "${INSTALL_ROOT}/Foo/Foo-0.2.3/lib/python2.8/site-packages"
                ),
                "requirements": [
                    "python >= 2.8, <2.9"
                ]
            }
        ]
    }


def test_update_without_version():
    """Update definition without version."""
    definition = wiz.definition.Definition({
        "identifier": "foo",
        "namespace": "plugin",
        "description": "This is a library",
    })

    mapping = {
        "identifier": "Foo-0.2.3",
        "name": "Foo",
        "key": "foo",
        "version": "0.2.3",
        "description": "This is a cool library",
        "target": "Foo/Foo-0.2.3",
        "python": {
            "identifier": "2.8",
            "request": "python >= 2.8, <2.9",
            "library-path": "lib/python2.8/site-packages"
        },
    }

    result = qip.definition.update(definition, mapping, "/packages")

    assert result.data() == {
        "identifier": "foo",
        "version": "0.2.3",
        "namespace": "plugin",
        "description": "This is a library",
        "install-root": "/packages",
        "environ": {
            "PYTHONPATH": "${INSTALL_LOCATION}:${PYTHONPATH}"
        },
        "variants": [
            {
                "identifier": "2.8",
                "install-location": (
                    "${INSTALL_ROOT}/Foo/Foo-0.2.3/lib/python2.8/site-packages"
                ),
                "requirements": [
                    "python >= 2.8, <2.9"
                ]
            }
        ]
    }


def test_update_without_namespace():
    """Update definition without namespace."""
    definition = wiz.definition.Definition({
        "identifier": "foo",
        "version": "0.2.3",
        "description": "This is a library",
    })

    mapping = {
        "identifier": "Foo-0.2.3",
        "name": "Foo",
        "key": "foo",
        "version": "0.2.3",
        "description": "This is a cool library",
        "target": "Foo/Foo-0.2.3",
        "python": {
            "identifier": "2.8",
            "request": "python >= 2.8, <2.9",
            "library-path": "lib/python2.8/site-packages"
        },
    }

    result = qip.definition.update(definition, mapping, "/packages")

    assert result.data() == {
        "identifier": "foo",
        "version": "0.2.3",
        "namespace": "library",
        "description": "This is a library",
        "install-root": "/packages",
        "environ": {
            "PYTHONPATH": "${INSTALL_LOCATION}:${PYTHONPATH}"
        },
        "variants": [
            {
                "identifier": "2.8",
                "install-location": (
                    "${INSTALL_ROOT}/Foo/Foo-0.2.3/lib/python2.8/site-packages"
                ),
                "requirements": [
                    "python >= 2.8, <2.9"
                ]
            }
        ]
    }


def test_update_with_system():
    """Update definition with system."""
    definition = wiz.definition.Definition({
        "identifier": "foo",
        "version": "0.2.3",
        "namespace": "plugin",
        "description": "This is a library",
    })

    mapping = {
        "identifier": "Foo-0.2.3",
        "name": "Foo",
        "key": "foo",
        "version": "0.2.3",
        "description": "This is a cool library",
        "target": "Foo/Foo-0.2.3",
        "python": {
            "identifier": "2.8",
            "request": "python >= 2.8, <2.9",
            "library-path": "lib/python2.8/site-packages"
        },
        "system": {
            "platform": "linux",
            "arch": "x86_64",
            "os": {
                "name": "centos",
                "major_version": 7
            }
        }
    }

    result = qip.definition.update(definition, mapping, "/packages")

    assert result.data() == {
        "identifier": "foo",
        "version": "0.2.3",
        "namespace": "plugin",
        "description": "This is a library",
        "install-root": "/packages",
        "environ": {
            "PYTHONPATH": "${INSTALL_LOCATION}:${PYTHONPATH}"
        },
        "system": {
            "platform": "linux",
            "arch": "x86_64",
            "os": "centos >= 7, < 8"
        },
        "variants": [
            {
                "identifier": "2.8",
                "install-location": (
                    "${INSTALL_ROOT}/Foo/Foo-0.2.3/lib/python2.8/site-packages"
                ),
                "requirements": [
                    "python >= 2.8, <2.9"
                ]
            }
        ]
    }


def test_update_with_commands():
    """Update definition with commands."""
    definition = wiz.definition.Definition({
        "identifier": "foo",
        "version": "0.2.3",
        "namespace": "plugin",
        "description": "This is a library",
    })

    mapping = {
        "target": "Foo/Foo-0.2.3",
        "python": {
            "identifier": "2.8",
            "request": "python >= 2.8, <2.9",
            "library-path": "lib/python2.8/site-packages"
        },
        "command": {
            "foo": "python -m foo"
        }
    }

    result = qip.definition.update(definition, mapping, "/packages")

    assert result.data() == {
        "identifier": "foo",
        "version": "0.2.3",
        "namespace": "plugin",
        "description": "This is a library",
        "install-root": "/packages",
        "command": {
            "foo": "python -m foo"
        },
        "environ": {
            "PYTHONPATH": "${INSTALL_LOCATION}:${PYTHONPATH}"
        },
        "variants": [
            {
                "identifier": "2.8",
                "install-location": (
                    "${INSTALL_ROOT}/Foo/Foo-0.2.3/lib/python2.8/site-packages"
                ),
                "requirements": [
                    "python >= 2.8, <2.9"
                ]
            }
        ]
    }


def test_update_with_commands_updated():
    """Update definition with commands updated."""
    definition = wiz.definition.Definition({
        "identifier": "foo",
        "version": "0.2.3",
        "namespace": "plugin",
        "description": "This is a library",
        "command": {
            "test": "Test",
            "foo": "FooX"
        }
    })

    mapping = {
        "target": "Foo/Foo-0.2.3",
        "python": {
            "identifier": "2.8",
            "request": "python >= 2.8, <2.9",
            "library-path": "lib/python2.8/site-packages"
        },
        "command": {
            "foo": "python -m foo"
        }
    }

    result = qip.definition.update(definition, mapping, "/packages")

    assert result.data() == {
        "identifier": "foo",
        "version": "0.2.3",
        "install-root": "/packages",
        "namespace": "plugin",
        "description": "This is a library",
        "command": {
            "test": "Test",
            "foo": "python -m foo"
        },
        "environ": {
            "PYTHONPATH": "${INSTALL_LOCATION}:${PYTHONPATH}"
        },
        "variants": [
            {
                "identifier": "2.8",
                "install-location": (
                    "${INSTALL_ROOT}/Foo/Foo-0.2.3/lib/python2.8/site-packages"
                ),
                "requirements": [
                    "python >= 2.8, <2.9"
                ]
            }
        ]
    }


def test_update_editable():
    """Update definition with mapping in editable mode."""
    definition = wiz.definition.Definition({
        "identifier": "foo",
        "version": "0.2.3",
        "namespace": "plugin",
        "description": "This is a library",
    })

    mapping = {
        "target": "Foo/Foo-0.2.3",
        "location": "/path/to/lib",
        "python": {
            "identifier": "2.8",
            "request": "python >= 2.8, <2.9",
            "library-path": "lib/python2.8/site-packages"
        },
    }

    result = qip.definition.update(
        definition, mapping, "/packages", editable_mode=True
    )

    assert result.data() == {
        "identifier": "foo",
        "version": "0.2.3",
        "namespace": "plugin",
        "description": "This is a library",
        "environ": {
            "PYTHONPATH": "${INSTALL_LOCATION}:${PYTHONPATH}"
        },
        "variants": [
            {
                "identifier": "2.8",
                "install-location": "/path/to/lib",
                "requirements": [
                    "python >= 2.8, <2.9"
                ]
            }
        ]
    }


def test_update_with_additional_variants_1():
    """Update definition with mapping with additional variants."""
    definition = wiz.definition.Definition({
        "identifier": "foo",
        "version": "0.2.3",
        "namespace": "plugin",
        "description": "This is a library",
    })

    mapping = {
        "target": "Foo/Foo-0.2.3",
        "python": {
            "identifier": "2.8",
            "request": "python >= 2.8, <2.9",
            "library-path": "lib/python2.8/site-packages"
        },
    }

    result = qip.definition.update(
        definition, mapping, "/packages",
        additional_variants=[
            {
                "identifier": "variant",
            },
            {
                "identifier": "3.6",
                "environ": {"KEY36": "VALUE36"}
            },
            {
                "identifier": "2.8",
                "environ": {"KEY28": "VALUE28"}
            },
            {
                "identifier": "2.7",
                "environ": {"KEY22": "VALUE22"}
            }
        ]
    )

    assert result.data() == {
        "identifier": "foo",
        "version": "0.2.3",
        "namespace": "plugin",
        "description": "This is a library",
        "install-root": "/packages",
        "environ": {
            "PYTHONPATH": "${INSTALL_LOCATION}:${PYTHONPATH}",
        },
        "variants": [
            {
                "identifier": "3.6",
                "environ": {"KEY36": "VALUE36"}
            },
            {
                "identifier": "2.8",
                "install-location": (
                    "${INSTALL_ROOT}/Foo/Foo-0.2.3/lib/python2.8/site-packages"
                ),
                "environ": {
                    "KEY28": "VALUE28"
                },
                "requirements": [
                    "python >= 2.8, <2.9"
                ]
            },
            {
                "identifier": "2.7",
                "environ": {"KEY22": "VALUE22"}
            },
            {
                "identifier": "variant",
            }
        ]
    }


def test_update_with_additional_variants_2():
    """Update definition with mapping with additional variants."""
    definition = wiz.definition.Definition({
        "identifier": "foo",
        "version": "0.2.3",
        "namespace": "plugin",
        "description": "This is a library",
    })

    mapping = {
        "target": "Foo/Foo-0.2.3",
        "python": {
            "identifier": "2.8",
            "request": "python >= 2.8, <2.9",
            "library-path": "lib/python2.8/site-packages"
        },
    }

    result = qip.definition.update(
        definition, mapping, "/packages",
        additional_variants=[
            {
                "identifier": "variant",
            },
            {
                "identifier": "3.6",
                "environ": {"KEY36": "VALUE36"}
            },
            {
                "identifier": "2.7",
                "environ": {"KEY22": "VALUE22"}
            }
        ]
    )

    assert result.data() == {
        "identifier": "foo",
        "version": "0.2.3",
        "namespace": "plugin",
        "description": "This is a library",
        "install-root": "/packages",
        "environ": {
            "PYTHONPATH": "${INSTALL_LOCATION}:${PYTHONPATH}"
        },
        "variants": [
            {
                "identifier": "3.6",
                "environ": {"KEY36": "VALUE36"}
            },
            {
                "identifier": "2.8",
                "install-location": (
                    "${INSTALL_ROOT}/Foo/Foo-0.2.3/lib/python2.8/site-packages"
                ),
                "requirements": [
                    "python >= 2.8, <2.9"
                ]
            },
            {
                "identifier": "2.7",
                "environ": {"KEY22": "VALUE22"}
            },
            {
                "identifier": "variant",
            }
        ]
    }


@pytest.mark.parametrize("identifier1, identifier2, expected", [
    ("2.7", "3.6", 1),
    ("3.6", "2.7", -1),
    ("abc", "def", -1),
    ("def", "abc", 1),
    ("2.7", "abc", -1),
    ("abc", "2.7", 1),
    ("2.7", "2.7", 0),
], ids=[
    "float",
    "float-inv",
    "string",
    "string-inv",
    "float-right",
    "float-left",
    "identical",
])
def test_compare_variants(identifier1, identifier2, expected):
    """Compare identifier values from variant mappings."""
    variant1 = {"identifier": identifier1}
    variant2 = {"identifier": identifier2}

    assert qip.definition._compare_variants(variant1, variant2) == expected

    # If result is -1, variant order stays the same.
    if expected == -1:
        assert sorted(
            [variant1, variant2],
            key=functools.cmp_to_key(qip.definition._compare_variants)
        ) == [variant1, variant2]

    # If result is 1, variant order is inverted.
    if expected == 1:
        assert sorted(
            [variant1, variant2],
            key=functools.cmp_to_key(qip.definition._compare_variants)
        ) == [variant2, variant1]



# :coding: utf-8

import platform as _platform

import distro
from packaging.version import Version, InvalidVersion


def query():
    """Return system mapping.

    :raise RuntimeError: if platform is not supported.

    :return: system mapping.

     It should be in the form of::

        {
            "platform": "linux",
            "arch": "x86_64",
            "os": {
                "name": "centos",
                "major_version": 7
            }
        }

    """
    name = _platform.system().lower()

    try:
        if name == "linux":
            mapping = query_linux()
        elif name == "darwin":
            mapping = query_mac()
        elif name == "windows":
            mapping = query_windows()
        else:
            raise RuntimeError

    except InvalidVersion as error:
        raise RuntimeError(
            "The operating system version found seems incorrect [{}]".format(
                error
            )
        )

    return mapping


def query_linux():
    """Return Linux system mapping.

    :return: system mapping.

    """
    distribution, version, _ = distro.linux_distribution(
        full_distribution_name=False
    )

    return {
        "platform": "linux",
        "arch": _platform.machine(),
        "os": {
            "name": distribution,
            "major_version": _extract_major_version(version)
        }
    }


def query_mac():
    """Return mac system mapping.

    :return: system mapping.

    """
    return {
        "platform": "mac",
        "arch": _platform.machine(),
        "os": {
            "name": "mac",
            "major_version": _extract_major_version(_platform.mac_ver()[0])
        }
    }


def query_windows():
    """Return windows system mapping.

    .. warning::

        The Windows versions superior to 8 will not be recognised properly with
        a Python version under 2.7.11

        https://hg.python.org/cpython/raw-file/53d30ab403f1/Misc/NEWS

        Also a bug as been introduced in Python 2.7.11 that prevent the
        recognition of old Windows version

        https://bugs.python.org/issue26513

    :return: system mapping.

    """
    return {
        "platform": "windows",
        "arch": _platform.machine(),
        "os": {
            "name": "windows",
            "major_version": _extract_major_version(_platform.win32_ver()[1])
        }
    }


def _extract_major_version(version):
    """Extract major version of operating system request for *version*.

    :param version: Version instance.

    :return: Major version integer.

    """
    version = Version(version)
    return int(version.base_version.split(".", 1)[0])



# :coding: utf-8

import logging
import json
import os
import tempfile

import wiz

import qip.command

#: Path to the python info script.
PYTHON_INFO_SCRIPT = os.path.join(
    os.path.dirname(__file__), "package_data", "python_info.py"
)


def fetch(python_target, mapping=None):
    """Fetch mapping with all environment variables required.

    :param python_target: Target a specific Python version via a Wiz request or
        a path to a Python executable (e.g. "python==2.7.*" or
        "/path/to/bin/python").

    :param mapping: optional custom environment mapping to be added to initial
        environment.

    :return: environment mapping

        It should be in the form of::

            {
                "PATH": "/path/to/bin",
                "PYTHONPATH": "/path/to/lib/site-packages",
            }

    Example::

        >>> fetch("python==2.7.*")
        >>> fetch("/path/to/bin/python")

    """
    logger = logging.getLogger(__name__ + ".fetch")
    logger.debug("initial environment: {}".format(mapping))

    if mapping is None:
        mapping = {}

    # If a Python executable is provided, use it instead of the Wiz request.
    if os.path.isfile(python_target) or os.sep in python_target:

        # Use symlink to executable in isolated new folder to ensure that
        # no other python version gets picked up.
        path = tempfile.mkdtemp(prefix="qip-env-")
        exec_name = os.path.basename(python_target)
        exec_path = os.path.join(path, exec_name)
        os.symlink(python_target, exec_path)

        # If executable is not named "python", create extra symlink.
        if exec_name != "python":
            os.symlink(exec_path, os.path.join(path, "python"))

        environ_mapping = mapping.copy()
        environ_mapping.update({"PATH": "{}:${{PATH}}".format(path)})
        context = {"environ": environ_mapping}

    else:
        context = wiz.resolve_context([python_target], environ_mapping=mapping)

    return context["environ"]


def fetch_python_mapping(environ_mapping):
    """Fetch Python version mapping.

    :param environ_mapping: mapping of environment variables

    :return: python mapping.

        It should be in the form of::

            {
                "identifier": "2.7",
                "request": "python >= 2.7, < 2.8",
                "installation-target": "lib/python2.7/site-packages"
            }

    """
    result = qip.command.execute(
        "python {}".format(PYTHON_INFO_SCRIPT), environ_mapping, quiet=True
    )
    try:
        mapping = json.loads(result)
    except ValueError:
        raise RuntimeError(
            "Impossible to fetch Python version mapping'"
        )

    return mapping



# :coding: utf-8

import logging
import functools
import os

import wiz
import wiz.environ
import wiz.exception
import wiz.symbol
import wiz.utility

#: Common namespace for all :term:`Wiz` definition.
NAMESPACE = "library"


def export(
    path, package_mapping, output_path, editable_mode=False,
    custom_definition=None, existing_definition=None
):
    """Export :term:`Wiz` definition to *path* for package mapping.

    :param path: destination path for the :term:`Wiz` definition.

    :param package_mapping: mapping of the python package built as returned by
        :func:`qip.package.install`.

    :param output_path: root destination path for Python packages installation.

    :param editable_mode: indicate whether the Python package location should
        target the source installation package. Default is False.

    :param custom_definition: :class:`wiz.definition.Definition` instance to
        update as returned by :func:`fetch_custom`. Default is None, which means
        that a default definition will be created from package mapping only.

    :param existing_definition: :class:`wiz.definition.Definition` instance to
        extract additional variants from as returned by :func:`fetch_existing`.
        Default is None which means that no additional variants will be added to
        new definition exported.

    """
    # Extract additional variants from existing definition if possible.
    additional_variants = None
    if existing_definition is not None:
        additional_variants = [
            variant.data() for variant in existing_definition.variants
        ]

    # Update definition or create a new definition.
    if custom_definition is not None:
        definition = update(
            custom_definition, package_mapping, output_path,
            editable_mode=editable_mode,
            additional_variants=additional_variants,
        )

    else:
        definition = create(
            package_mapping, output_path,
            editable_mode=editable_mode,
            additional_variants=additional_variants
        )

    wiz.export_definition(path, definition, overwrite=True)


def fetch_custom(package_mapping):
    """Retrieve :term:`Wiz` definition from package mapping installed.

    Return the :term:`Wiz` definition extracted from a
    :file:`package_data/wiz.json` file found within the package installation
    path.

    :param package_mapping: mapping of the python package built as returned by
        :func:`qip.package.install`.

    :raise wiz.exception.WizError: if the :term:`Wiz` definition found is
        incorrect.

    :return: :class:`wiz.definition.Definition` instance fetched, or None if
        no definition was found.

    .. seealso:: :ref:`development/custom_definition`

    """
    logger = logging.getLogger(__name__ + ".retrieve")

    definition_path = os.path.join(
        package_mapping["location"], package_mapping["module_name"],
        "package_data", "wiz.json"
    )

    if os.path.exists(definition_path):
        definition = wiz.load_definition(definition_path)
        logger.info(
            "\tWiz definition extracted from '{}'.".format(
                package_mapping["identifier"]
            )
        )
        return definition


def fetch_existing(package_mapping, definition_mapping, namespace=None):
    """Retrieve corresponding :term:`Wiz` definition in definition mapping.

    :param package_mapping: mapping of the python package built as returned by
        :func:`qip.package.install`.

    :param definition_mapping: mapping regrouping all available definitions as
        returned by :func:`wiz.fetch_definition_mapping`.

    :param namespace: Namespace of the definition to fetch. Default is
        :data:`NAMESPACE`.

    :return: :class:`wiz.definition.Definition` instance fetched, or None if
        no definition was found.

    """
    namespace = namespace or NAMESPACE

    try:
        return wiz.fetch_definition(
            "{0}::{1[key]}=={1[version]}".format(namespace, package_mapping),
            definition_mapping
        )
    except wiz.exception.RequestNotFound:
        pass


def create(
    package_mapping, output_path, editable_mode=False, additional_variants=None
):
    """Create :term:`Wiz` definition from package mapping.

    :param package_mapping: mapping of the python package built as returned by
        :func:`qip.package.install`.

    :param output_path: root destination path for Python packages installation.

    :param editable_mode: indicate whether the Python package location should
        target the source installation package. Default is False.

    :param additional_variants: None or list of variant mappings that should be
        added to the definition created. Default is None.

    :return: :class:`wiz.definition.Definition` instance created.

    """
    logger = logging.getLogger(__name__ + ".create")

    definition_data = {
        "identifier": package_mapping["key"],
        "version": package_mapping["version"],
        "description": package_mapping["description"],
        "namespace": NAMESPACE,
        "environ": {
            "PYTHONPATH": "${{{}}}:${{PYTHONPATH}}".format(
                wiz.symbol.INSTALL_LOCATION
            )
        }
    }

    # Add commands mapping.
    if "command" in package_mapping.keys():
        definition_data["command"] = package_mapping["command"]

    # Add system constraint if necessary.
    if "system" in package_mapping.keys():
        definition_data["system"] = _process_system_mapping(package_mapping)

    # Target package location if the installation is in editable mode.
    location_path = package_mapping.get("location", "")

    if not editable_mode:
        definition_data["install-root"] = output_path
        location_path = os.path.join(
            "${{{}}}".format(wiz.symbol.INSTALL_ROOT),
            package_mapping["target"],
            package_mapping["python"]["library-path"]
        )

    # Update and set variant for python version.
    variants = []

    if additional_variants is not None:
        variants = sorted(
            additional_variants,
            key=functools.cmp_to_key(_compare_variants)
        )

    _update_variants(variants, package_mapping, location_path)

    definition_data["variants"] = variants

    definition = wiz.definition.Definition(definition_data)
    logger.info(
        "\tWiz definition created for '{0[identifier]}'.".format(
            package_mapping
        )
    )
    return definition


def update(
    definition, package_mapping, output_path, editable_mode=False,
    additional_variants=None
):
    """Update *definition* from package mapping.

    :param definition: :class:`wiz.definition.Definition` instance as returned
        by :func:`fetch_custom`.

    :param package_mapping: mapping of the python package built as returned by
        :func:`qip.package.install`.

    :param output_path: root destination path for Python packages installation.

    :param editable_mode: indicate whether the Python package location should
        target the source installation package. Default is False.

    :param additional_variants: None or list of variant mappings that should be
        added to the definition updated. Default is None.

    :return: Updated :class:`wiz.definition.Definition` instance.

    """
    if not definition.description:
        definition = definition.set(
            "description", package_mapping["description"]
        )

    if not definition.version:
        definition = definition.set("version", package_mapping["version"])

    if not definition.namespace:
        definition = definition.set("namespace", NAMESPACE)

    if not definition.system and package_mapping.get("system"):
        definition = definition.set(
            "system", _process_system_mapping(package_mapping)
        )

    if package_mapping.get("command"):
        definition = definition.update("command", package_mapping["command"])

    # Update environ mapping
    environ_mapping = {
        "PYTHONPATH": "${{{}}}:${{PYTHONPATH}}".format(
            wiz.symbol.INSTALL_LOCATION
        )
    }

    python_path = definition.environ.get("PYTHONPATH")
    if python_path:
        environ_mapping["PYTHONPATH"] = wiz.environ.substitute(
            environ_mapping["PYTHONPATH"], {"PYTHONPATH": python_path}
        )

    definition = definition.update("environ", environ_mapping)

    # Target package location if the installation is in editable mode.
    package_path = package_mapping.get("location", "")

    if not editable_mode:
        definition = definition.set("install-root", output_path)
        package_path = os.path.join(
            "${{{}}}".format(wiz.symbol.INSTALL_ROOT),
            package_mapping["target"],
            package_mapping["python"]["library-path"]
        )

    variants = definition.variants

    if additional_variants is not None:
        variants = sorted(
            variants + additional_variants,
            key=functools.cmp_to_key(_compare_variants)
        )

    _update_variants(variants, package_mapping, package_path)

    return definition.set("variants", variants)


def _update_variants(variants, package_mapping, path):
    """Add variant corresponding to *identifier* to the *variant* list.

    Update existing variant if necessary or add new variant corresponding to the
    python version required. If a new variant is added, it will be inserted
    to the variant list so that the highest Python version is always first.

    :param variants: list of variant mappings to update.

    :param package_mapping: mapping of the python package built as returned by
        :func:`qip.package.install`.

    :param path: path where python package has been installed.

    :return: None.

    .. note::

        The *variants* list will be mutated.

    """
    identifier = package_mapping["python"]["identifier"]
    python_request = package_mapping["python"]["request"]

    # Process all requirements to detect duplication.
    requirements = _process_requirements(package_mapping, python_request)

    # Index of new variant if necessary.
    _index = 0

    for index, variant in enumerate(variants):
        if variant["identifier"] != identifier:

            # Update index for new variant.
            if _compare_variants({"identifier": identifier}, variant) > 0:
                _index = index + 1

            continue

        variant["install-location"] = path

        # Add requirements that are not already in the definition.
        variant.setdefault("requirements", [])
        variant["requirements"] += [
            req for req in requirements
            if not any(
                req.replace(" ", "") == _req.replace(" ", "")
                for _req in variant["requirements"]
            )
        ]

        del variants[index]
        variants.insert(index, variant)
        return

    # If no variant has been updated, create a new variant.
    variant = {
        "identifier": identifier,
        "install-location": path,
        "requirements": requirements
    }

    variants.insert(_index, variant)


def _compare_variants(variant1, variant2):
    """Compare identifier values from variant mappings.

    Both identifiers will be converted into a negative float if possible (e.g.
    "2.7" will become -2.7). If one or both identifiers cannot be converted, the
    string value  is kept.

    If both identifiers are of the same type:

    * Return -1 if *identifier2* if higher than *identifier1*.
    * Return 1 if *identifier1* if higher than *identifier2*.
    * Return 0 if *identifier1* if higher than *identifier2*.

    If only *identifier1* is converted into a negative float, -1 is returned.

    If only *identifier2* is converted into a negative float, 1 is returned.

    :param variant1: Variant reference mapping.

    :param variant2: Variant reference mapping to compare *variant1* with.

    :return: Numerical value following the rules above (-1, 1 or 0).

    """
    try:
        identifier1 = -float(variant1["identifier"])
    except ValueError:
        identifier1 = variant1["identifier"]

    try:
        identifier2 = -float(variant2["identifier"])
    except ValueError:
        identifier2 = variant2["identifier"]

    if type(identifier1) == type(identifier2):
        if identifier1 == identifier2:
            return 0
        return -1 if identifier1 < identifier2 else 1

    elif isinstance(identifier1, float):
        return -1
    return 1


def _process_system_mapping(package_mapping):
    """Compute 'system' keyword for the :term:`Wiz` definition from mapping.

    :param package_mapping: mapping of the python package built as returned by
        :func:`qip.package.install`.

    :return: system mapping.

    """
    major_version = package_mapping["system"]["os"]["major_version"]
    return {
        "platform": package_mapping["system"]["platform"],
        "arch": package_mapping["system"]["arch"],
        "os": (
            "{name} >= {min_version}, < {max_version}".format(
                name=package_mapping["system"]["os"]["name"],
                min_version=major_version,
                max_version=major_version + 1,
            )
        )
    }


def _process_requirements(package_mapping, python_request):
    """Compute 'requirements' keyword for the :term:`Wiz` definition.

    :param package_mapping: mapping of the python package built as returned by
        :func:`qip.package.install`.

    :param python_request: Python version requirement (e.g.
        "python >=2.7, <2.8")

    :return: requirements list.

    """
    requests = [python_request]

    # Add the library namespace for all requirements fetched.
    for request in package_mapping.get("requirements", []):
        request = "{}::{}".format(NAMESPACE, request)
        requirement = wiz.utility.get_requirement(request)
        requirement.extras = {package_mapping["python"]["identifier"]}
        requests.append(str(requirement))

    return requests
